from __future__ import annotations

from typing import Protocol, TypeVar, Tuple, Any, Dict

from eventz.messages import Event, Command

T = TypeVar("T")
Events = Tuple[Event, ...]


class ProcessesCommandsProtocol(Protocol):  # pragma: no cover
    def process(self, command: Command) -> Events:
        ...


class RepositoryProtocol(Protocol[T]):  # pragma: no cover
    def create(self, **kwargs) -> Events:
        ...

    def read(self, aggregate_id: str) -> T:
        ...

    def persist(self, aggregate_id: str, events: Events) -> None:
        ...


class AggregateBuilderProtocol(Protocol[T]):  # pragma: no cover
    def create(self, events: Events) -> T:
        ...

    def update(self, aggregate: T, events: Events) -> T:
        ...


class MarshallProtocol(Protocol):  # pragma: no cover
    def to_json(self, obj: Any) -> str:
        ...

    def from_json(self, json_string: str) -> Any:
        ...

    def register_codec(self, name: str, handler: MarshallCodecProtocol):
        ...

    def deregister_codec(self, name: str):
        ...


class MarshallCodecProtocol(Protocol):  # pragma: no cover
    def serialise(self, obj: Any) -> Dict:
        ...

    def deserialise(self, params: Dict) -> Any:
        ...

    def handles(self, obj: Any) -> bool:
        ...


class JsonSerlialisable(Protocol):  # pragma: no cover
    def get_json_data(self) -> Dict:
        ...


class EventStoreProtocol(Protocol):  # pragma: no cover
    def fetch(self, aggregate_id: str) -> Events:
        ...

    def persist(self, aggregate_id: str, events: Events) -> None:
        ...


from __future__ import annotations

from typing import Optional

from eventz.entity import Entity


class Aggregate(Entity):
    """
    Top-level entity that other services interact with from outside the domain.
    """

    def __init__(self, uuid: Optional[str] = None):
        super().__init__(uuid)

    @staticmethod
    def make_id() -> str:
        return Entity.make_id()



from typing import TypeVar, Any, Dict

T = TypeVar("T")


class Immutable(type):
    def __new__(mcs, name, bases, dictionary):
        instance = type.__new__(mcs, name, bases, dictionary)
        instance.__setattr__ = _setattr
        instance.__delattr__ = _delattr
        instance.__immutable__ = False
        return instance

    def __call__(cls, *args, **kwargs):
        """
        Allow attributes to be set once on instantiation
        """
        if "__immutable__" in kwargs:
            del kwargs["__immutable__"]
        obj = type.__call__(cls, *args, **kwargs)
        # now instance is instantiated set as immutable
        obj.__immutable__ = True
        return obj

    @staticmethod
    def __mutate__(
        obj: T, name: str, value: Any, transform_underscores: bool = False
    ) -> T:
        """
        Mutates a single parameter of an object and returns a new, updated copy
        If `transform_underscores` is set to True then
        private attributes set as `_attr` will be passed to `__init__` as `attr`
        """
        # obtain a dictionary of the objects attributes
        attrs: Dict[str, Any] = {k: getattr(obj, k) for k in vars(obj)}
        # set the mutated attribute to the new value
        attrs[name] = value
        if transform_underscores:
            new_attrs = {}
            # remove underscore from the start of private properties
            for attr_name, attr_val in attrs.items():
                if attr_name.startswith("_") and attr_name != "__immutable__":
                    attr_name = attr_name[1:]
                new_attrs[attr_name] = attr_val
            attrs = new_attrs
        # return a new instance of the object
        return type(obj)(**attrs)


def _setattr(self, attr, value):
    # we'll only block "setting" once __immutable__ is True
    if self.__immutable__:
        raise AttributeError("An attribute cannot be modified on an immutable object.")
    object.__setattr__(self, attr, value)


def _delattr(self, attr):  # pragma: no cover
    raise AttributeError("An attribute cannot be deleted on an immutable object.")



from typing import Dict

from eventz.protocols import Events, EventStoreProtocol


class DummyStorage(EventStoreProtocol):
    def __init__(self):
        self._persisted_events: Dict[str, Events] = {}
        self._fetch_called: int = 0

    def fetch(self, aggregate_id: str) -> Events:
        self._fetch_called += 1
        return self._persisted_events[aggregate_id]

    def persist(self, aggregate_id: str, events: Events):
        if aggregate_id in self._persisted_events:
            self._persisted_events[aggregate_id] = (
                self._persisted_events[aggregate_id] + events
            )
        else:
            self._persisted_events[aggregate_id] = events

    @property
    def fetch_called(self) -> int:
        return self._fetch_called



from typing import Optional, Dict, Any, Generic, TypeVar
from uuid import uuid4

from eventz.immutable import Immutable

T = TypeVar("T")


class Entity(Generic[T], metaclass=Immutable):
    transform_underscores: bool = False

    @staticmethod
    def make_id() -> str:
        return str(uuid4())

    def __init__(self, uuid: Optional[str] = None):
        self.uuid: str = uuid if uuid is not None else self.make_id()

    def __eq__(self, other) -> bool:
        if not isinstance(other, Entity):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        if not isinstance(other, Entity):
            return True
        return self.__dict__ != other.__dict__

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        attrs = {k: getattr(self, k) for k in vars(self)}
        attrs_string = " ".join([f"{k}={v}" for k, v in attrs.items()])
        return f"{class_name}({attrs_string})"

    def _mutate(self, name, value) -> T:
        return Immutable.__mutate__(
            self, name, value, self.transform_underscores
        )

    def _mutate_all(self, updates: Dict[str, Any]) -> T:
        entity = self
        for name, value in updates.items():
            entity = Immutable.__mutate__(
                entity, name, value, self.transform_underscores
            )
        return entity



from typing import TypeVar

from eventz.aggregate import Aggregate
from eventz.protocols import (
    RepositoryProtocol,
    AggregateBuilderProtocol,
    Events,
    EventStoreProtocol,
)

T = TypeVar("T")


class Repository(RepositoryProtocol[T]):
    def __init__(
        self,
        aggregate_class: type,
        storage: EventStoreProtocol,
        builder: AggregateBuilderProtocol,
    ):
        self._aggregate_class: type = aggregate_class
        self._storage: EventStoreProtocol = storage
        self._builder: AggregateBuilderProtocol = builder

    def create(self, **kwargs) -> Events:
        if "uuid" not in kwargs:
            kwargs["uuid"] = Aggregate.make_id()
        events = getattr(self._aggregate_class, "create")(**kwargs)
        self._storage.persist(kwargs["uuid"], events)
        return events

    def read(self, aggregate_id: str) -> T:
        events = self._storage.fetch(aggregate_id=aggregate_id)
        return self._builder.create(events)

    def persist(self, aggregate_id: str, events: Events) -> None:
        self._storage.persist(aggregate_id, events)



from datetime import datetime
from uuid import uuid4

from eventz.value_object import ValueObject


class Message(ValueObject):
    def __init__(self, uuid: str = None, timestamp: datetime = None):
        try:
            getattr(self, "version")
        except AttributeError:
            err = (
                "Child classes of Message cannot be instantiated "
                "without a 'version' attribute set on the class."
            )
            raise TypeError(err)
        self.uuid: str = uuid or str(uuid4())
        self.timestamp: datetime = timestamp or datetime.utcnow()


class Event(Message):
    pass


class Command(Message):
    pass



from abc import ABC, abstractmethod
from typing import Dict, TypeVar

from eventz.aggregate import Aggregate
from eventz.messages import Event
from eventz.protocols import AggregateBuilderProtocol, Events

T = TypeVar("T")


class AggregateBuilder(ABC, AggregateBuilderProtocol):
    def create(self, events: Events) -> T:
        kwargs = {}
        return self._apply_events(kwargs, events)

    def update(self, aggregate: Aggregate, events: Events) -> T:
        kwargs = vars(aggregate)
        return self._apply_events(kwargs, events)

    def _apply_events(self, kwargs: Dict, events: Events) -> T:
        for event in events:
            kwargs = self._apply_event(kwargs, event)
        return self._new_aggregate(kwargs)

    @abstractmethod
    def _apply_event(self, kwargs: Dict, event: Event) -> Dict:  # pragma: no cover
        raise NotImplementedError

    @abstractmethod
    def _new_aggregate(self, kwargs: Dict) -> T:  # pragma: no cover
        """
        i.e.: return MyAggregate(**kwargs)
        """
        raise NotImplementedError



import json
import os
import shutil
from typing import Sequence, Tuple

from eventz.event_store import EventStore
from eventz.messages import Event
from eventz.protocols import MarshallProtocol, EventStoreProtocol


class EventStoreJsonFile(EventStore, EventStoreProtocol):
    def __init__(
        self,
        storage_path: str,
        marshall: MarshallProtocol,
        recreate_storage: bool = True,
    ):
        self._storage_path: str = storage_path
        self._marshall = marshall
        if recreate_storage and os.path.isdir(self._storage_path):
            shutil.rmtree(self._storage_path)
            os.mkdir(self._storage_path)
            # toy/example implementation, so don't worry about security
            os.chmod(self._storage_path, 0o777)

    def fetch(self, aggregate_id: str) -> Tuple[Event, ...]:
        file_path = self._get_file_path(aggregate_id)
        if not os.path.isfile(file_path):
            return ()
        with open(file_path) as json_file:
            json_string = json_file.read()
        return tuple(self._marshall.from_json(json_string))

    def persist(self, aggregate_id: str, events: Sequence[Event]):
        if not os.path.isdir(self._storage_path):
            os.mkdir(self._storage_path)
        with open(f"{self._storage_path}/{aggregate_id}.json", "w+") as json_file:
            json.dump(self._marshall.to_json(events), json_file)

    def _get_file_path(self, aggregate_id: str) -> str:
        return f"{self._storage_path}/{aggregate_id}.json"



from abc import ABC, abstractmethod

from eventz.messages import Command
from eventz.protocols import ProcessesCommandsProtocol, RepositoryProtocol, Events


class Service(ABC, ProcessesCommandsProtocol):
    def __init__(self, repository: RepositoryProtocol):
        self._repository: RepositoryProtocol = repository

    @abstractmethod
    def process(self, command: Command) -> Events:
        raise NotImplementedError  # pragma: no cover



from typing import Generic, TypeVar

from eventz.immutable import Immutable

T = TypeVar("T")


class ValueObject(Generic[T], metaclass=Immutable):
    transform_underscores: bool = False

    def __eq__(self, other) -> bool:
        if not isinstance(other, ValueObject):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        if not isinstance(other, ValueObject):
            return True
        return self.__dict__ != other.__dict__

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        attrs = {k: getattr(self, k) for k in vars(self)}
        if "__immutable__" in attrs:
            del attrs["__immutable__"]
        attrs_string = " ".join([f"{k}={v}" for k, v in attrs.items()])
        return f"{class_name}({attrs_string})"

    def _mutate(self, name, value) -> T:
        return Immutable.__mutate__(
            self, name, value, self.transform_underscores
        )



import datetime
import importlib
import json
from enum import Enum
from typing import Any, Dict

import immutables
from dateutil.parser import parse

from eventz.protocols import MarshallCodecProtocol, MarshallProtocol


class Marshall(MarshallProtocol):
    def __init__(self, codecs: Dict[str, MarshallCodecProtocol] = None):
        self._codecs = {} if codecs is None else codecs

    def register_codec(self, fcn: str, codec: MarshallCodecProtocol):
        self._codecs[fcn] = codec

    def deregister_codec(self, fcn: str):
        del self._codecs[fcn]

    def has_codec(self, fcn: str):
        return fcn in self._codecs

    def to_json(self, data: Any) -> str:
        data = self._serialise_data(data)
        return json.dumps(data, sort_keys=True)

    def from_json(self, json_string: str) -> Any:
        data = json.loads(json_string)
        return self._deserialise_data(data)

    def _serialise_data(self, data: Any) -> Any:
        if self._is_handled_by_codec(data):
            return self._object_to_codec_dict(data)
        elif self._is_sequence(data):
            new_sequence = []
            for item in data:
                new_sequence.append(self._serialise_data(item))
            return new_sequence
        elif self._is_mapping(data):
            new_mapping = {}
            for key, value in data.items():
                new_mapping[key] = self._serialise_data(value)
            return new_mapping
        elif self._is_simple_type(data):
            return data
        else:
            return self._object_to_dict(data)

    def _deserialise_data(self, data: Any) -> Any:
        if self._is_enum_dict(data):
            return self._dict_to_enum(data)
        if self._is_serialised_class(data):
            return self._dict_to_object(data)
        elif self._requires_codec(data):
            return self._codec_dict_to_object(data)
        elif self._is_sequence(data):
            new_sequence = []
            for item in data:
                new_sequence.append(self._deserialise_data(item))
            return new_sequence
        elif self._is_mapping(data):
            new_mapping = {}
            for key, value in data.items():
                new_mapping[key] = self._deserialise_data(value)
            return immutables.Map(new_mapping)
        else:  # all other simple types now
            return data

    def _object_to_dict(self, obj: Any) -> Dict:
        data = {"__fcn__": self._get_fcn(obj)}
        if hasattr(obj, "version"):
            data["__version__"] = obj.version
        if hasattr(obj, "get_json_data") and callable(obj.get_json_data):
            json_data = obj.get_json_data()
        else:
            json_data = vars(obj)
        for attr, value in json_data.items():
            if not attr.startswith("__"):
                data[attr] = self._serialise_data(value)
        return data

    def _dict_to_object(self, data: Dict) -> Any:
        kwargs = {}
        for key, value in data.items():
            if not key.startswith("__"):
                kwargs[key] = self._deserialise_data(value)
        fully_qualified_name = data["__fcn__"]
        # @TODO add "allowed_namespaces" list to class and do a check here to protect against code injection
        _class = self._get_class_from_fcn(fully_qualified_name)
        return _class(**kwargs)

    def _codec_dict_to_object(self, data: Dict) -> Any:
        fcn = data["__codec__"]
        return self._codecs[fcn].deserialise(data["params"])

    def _object_to_codec_dict(self, obj: Any) -> Dict:
        for codec in self._codecs.values():
            if codec.handles(obj):
                return codec.serialise(obj)

    def _dict_to_enum(self, data: Dict) -> Enum:
        fully_qualified_name = data["__fcn__"]
        # @TODO add "allowed_namespaces" list to class and do a check here to protect against code injection
        _class = self._get_class_from_fcn(fully_qualified_name)
        return getattr(_class, data["_name_"])

    def _is_handled_by_codec(self, data: Any) -> bool:
        return any([codec.handles(data) for codec in self._codecs.values()])

    def _is_sequence(self, data: Any) -> bool:
        return isinstance(data, (list, tuple))

    def _is_mapping(self, data: Any) -> bool:
        return isinstance(data, (dict, set, immutables.Map))

    def _is_enum_dict(self, data: Dict) -> bool:
        return isinstance(data, Dict) and "_value_" in data and "_name_" in data

    def _is_simple_type(self, data: Any) -> bool:
        if type(data).__module__ == "builtins":
            return True
        # now check for any other types we want to treat as simple
        return isinstance(data, (datetime.datetime,))

    def _is_serialised_class(self, data: Any) -> bool:
        return isinstance(data, dict) and "__fcn__" in data

    def _requires_codec(self, data: Any) -> bool:
        return isinstance(data, dict) and "__codec__" in data

    def _get_fcn(self, obj):
        return obj.__class__.__module__ + "." + obj.__class__.__name__

    def _get_class_from_fcn(self, fcn: str) -> type:
        module_name, class_name = fcn.rsplit(".", 1)
        return getattr(importlib.import_module(module_name), class_name)


class NoCodecError(Exception):
    pass


class DatetimeCodec(MarshallCodecProtocol):
    def serialise(self, obj: Any) -> Dict:
        if not isinstance(obj, datetime.datetime):
            err = (
                "Only objects of type 'datetime.datetime' "
                "can be handled by DatetimeHandler codec."
            )
            raise TypeError(err)
        return {
            "__codec__": "eventz.marshall.DatetimeCodec",
            "params": {"timestamp": obj.isoformat()},
        }

    def deserialise(self, params: Dict) -> Any:
        return parse(params["timestamp"])

    def handles(self, obj: Any) -> bool:
        return isinstance(obj, datetime.datetime)



from typing import TypeVar, Generic, Iterable, Tuple

T = TypeVar('T')


class ImmutableSequence(Generic[T]):
    def __init__(self, items: Iterable[T]):
        self._items: Tuple[T] = tuple(items)

    def __eq__(self, other) -> bool:
        if not isinstance(other, ImmutableSequence):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        if not isinstance(other, ImmutableSequence):
            return True
        return self.__dict__ != other.__dict__

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        attrs = {k: getattr(self, k) for k in vars(self)}
        if "__immutable__" in attrs:  # pragma: no cover
            del attrs["__immutable__"]
        attrs_string = " ".join([f"{k}={v}" for k, v in attrs.items()])
        return f"{class_name}({attrs_string})"

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self):
        for item in self._items:
            yield item

    def __getitem__(self, key) -> T:
        return self._items[key]
